(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* cdsa definitions                                            *)
(*     (adapted from Yu's and Xiang's code)                    *)
(*                                                             *)
(***************************************************************)

(* $Id: cdsa.v,v 1.5 2006/07/26 07:48:27 ssg05 Exp $ *)

(* proof of the dynamic memory management system *)
Require Import machine.
Require Export Coq.Lists.List.
Require Export natpm.
Require Export seplogic.

Set Implicit Arguments.
Unset Strict Implicit.

(* glob consts *)
Notation NULL := O.
Notation HDSIZE := 2.
(* Parameter FLIST : nat. *)
(* Parameter TOP : nat. *)
Notation NXT := 0. 
Notation SZ := 1.

Notation prev := s1.
Notation p_ptr := s2.

Open Local Scope seplogic_scope.

(* definitions *)
Definition buffer (x n : nat) :=
  lexists l, !(length l = n) ** x |->, l.

(** _______p_____________________p+n-3_  **)
(**| q | s |                       |   | **)
(** -----------------------------------  **)
Definition MBlk (p q n : nat) : hpred :=
  !(p > 2 /\ n > 2) ** (p - 2 |-> q) ** (p - 1 |-> n) ** buffer p (n - 2).

(** p________   p'________          __________  **)
(** |p'|*|   |   | | |    | .......|q|        | **)
(**  --------     --------          ----------  **)
Fixpoint MBlkLst (n p q : nat) {struct n} : hpred :=
  match n with
    | O => !(p = q)
    | S n' => lexists p', !(p < p' \/ p' = NULL) **
      (lexists len, MBlk (p + 2) p' len) ** MBlkLst n' p' q
  end.

(** fl_________         __________     p_________  **)
(**  |  |  |   | ..... |p |       |    |q |      | **)
(**   ---------         ----------      ---------  **)
Definition EndL (fl p q : nat) :=
  if beq_nat p NULL then MBlkLst O fl q
    else !(p < q \/ q = NULL) ** 
      (lexists n, MBlkLst n fl p) ** (lexists len, MBlk (p + 2) q len).

(** fl_________         __________     p_________       __________   **)
(**  |  |  |   | ..... |p |       |    |q |      |...  |nil|       | **)
(**   ---------         ----------      ---------       -----------  **)
Definition MidL (fl p q : nat) :=
  EndL fl p q ** (lexists n, MBlkLst n q NULL).

(* good free list is a MBlkLst *)
Definition Good (fl : nat) :=
  lexists n, MBlkLst n fl NULL.

(* useful lemmas *)
Lemma buffer_add : forall n h x v A,
  h ||- buffer x n ** x + n |-> v ** A ->
    h ||- buffer x (S n) ** A.
Proof.
  induction n; intros.
  unfold buffer in * .
  sepelim H.
  destruct x0.
  lexintro 1; exists (v::nil).
  generalize H; simpl; simplpm.
  clear H; intro; simplsep 3 3.
  apply (unit_elim_right H).
  discriminate.
  unfold buffer in *.
  simpl in H.
  sepelim H.
  destruct x0.
  discriminate.
  destruct x0.
  assert (n = 0).
  simpl in A0;  omega.
  subst n.
  simpl in H.
  lexintro 1; exists (n0::v::nil).
  simplpm2h H.
  simpl; simplsep 3 3.

  simpl in H.
  assert (h ||- (lexists l, !(length l = S n) ** 
    S x |->, l) ** A ** (x |-> n0)).
  apply IHn with (v:= v).
  lexintro 1; exists (n1::x0).
  simplsep 4 4.
  replace (S x + n) with (x + S n); [exact H |  omega].
  sepelim H0; clear H.
  lexintro 1; exists (n0::x1).
  destruct x1.
  discriminate.
  simpl; simplsep 4 4.
Qed.
 
Lemma buffer_append : forall m h x n A,
  h ||- buffer x n ** buffer (x + n) m ** A ->
    h ||- buffer x (n + m) ** A.
Proof.
  induction m; intros.
  unfold buffer in * .
  sepelim H.
  destruct x1.
  simpl in H.
  lexintro 1; exists x0.
  simplsep 3 3.
  simplpm; trivial.
  apply (unit_elim_left H).
  discriminate.

  unfold buffer in * .
  sepelim H.
  destruct x1.
  discriminate.
  destruct x1.
  assert (m = 0).
  simpl in A1; omega.
  subst m.
  simpl in H.
  simplpm.
  assert (h ||- buffer x (S n) ** A).
  apply buffer_add with n0.
  unfold buffer.
  lexintro 1; exists x0; simplsep 4 4.
  apply H0.
  simpl in H.
  rewrite pSSp.
  apply IHm.
  assert (h ||- buffer x (S n) ** buffer (x + S n) m ** A).
  apply buffer_add with n0.
  unfold buffer.
  lexintro 1; exists x0.
  lexintro 4; exists (n1::x1).
  simplsep 5 5.
  simpl.
  rewrite pSSp.
  apply H.
  apply H0.
Qed.

Lemma buffer_elim : forall n h x y,
  n > 0 -> x <= y < x + n ->
  h ||- buffer x n ** top -> 
    exists v, h ||- y |-> v ** top.
Proof.
  destruct n as [ | n].
  intros; elimtype False;  omega.
  intros h x y _ R Hp.
  generalize dependent x.
  generalize h y; clear h y.
  induction n; intros;
    unfold buffer in Hp; sepelim Hp.
  destruct x0.
  discriminate.
  destruct x0.
  unfold contL in Hp.
  replace y with x.
  exists n; simplsep 4 4.
  omega.
  discriminate.
  compare y x; intro E; subst.
  destruct x0.
  discriminate.
  destruct x0.
  discriminate.
  simpl in Hp.
  exists n0; simplsep 4 4.
  destruct x0.
  discriminate.
  destruct x0.
  discriminate.
  apply IHn with (x := S x).
  omega.
  unfold buffer.
  lexintro 1; exists (n1::x0).
  simpl in * .
  simplsep 4 4.
Qed.

Lemma buffer_elim2 : forall n h x y v,
  n > 0 -> 
  h ||- buffer x n ** y |-> v ** top -> 
    y < x \/ y >= x + n.
Proof.
  intros.
  destruct H0 as [h1 [h2 [J [Hp1 Hp2]]]].
  assert (~ x <= y < x + n);
    [intro F |  omega].
  destruct (buffer_elim H F (top_intro_right Hp1)).
  hypreplace J (J y).
  rewrite (cont_get_eq1 Hp2) in J.
  rewrite (cont_get_eq1 H0) in J.
  destruct (Heap.get h y); inversion J.
Qed.

Lemma buffer_remove : forall n h x A,
  h ||- buffer x (S n) ** A ->
    exists y, h ||- buffer x n ** (x + n) |-> y ** A.
Proof.
  induction n; unfold buffer; simpl; intros.
  sepelim H.
  destruct x0.
  discriminate.
  destruct x0.
  simpl in H.
  exists n.
  lexintro 1; exists (nil (A := nat)).
  simplpm; simpl; simplsep 2 1.
  discriminate.
  sepelim H.
  destruct x0.
  discriminate.
  destruct x0.
  discriminate.
  assert (exists y, h ||- buffer (S x) n ** 
    (S x) + n |-> y ** A ** x |-> n0).
  apply IHn.
  simpl in H; simplsep 2 2.
  unfold buffer.
  lexintro 1; exists (n1::x0).
  simpl; simplsep 2 2.
  destruct H0.
  exists x1.
  clear H; unfold buffer in H0; sepelim H0.
  lexintro 1; exists (n0::x2).
  rewrite pSSp.
  simpl.
  simplsep 4 4.
  destruct x2; simpl in * ;simplsep 2 2.
  apply (unit_elim_right H0).
Qed.

Lemma buffer_split : forall m h x n A,
  h ||- buffer x n ** A -> n > m + 2 ->
    exists y, exists z,
      h ||- buffer x (n - (m + 2)) ** (x + (n - (m + 2))) |-> y **
	(S (x + (n - (m + 2)))) |-> z ** buffer (x + (n - m)) m ** A.
Proof.
  induction m; simpl; intros.
  sepelim H.
  destruct n as [ | n]; [elimtype False;  omega | idtac].
  destruct n as [ | n]; [elimtype False;  omega | idtac].
  destruct (buffer_remove H) as [y H']; clear H.
  destruct (buffer_remove H') as [z H]; clear H'.
  replace (S (S n) - 2) with n;
    [exists z; exists y |  omega].
  replace (S (x + n)) with (x + S n);
    [simplsep 4 4 |  omega].
  unfold buffer.
  lexintro 1; exists (nil (A:=nat)).
  simpl; simplsep 2 2.
  destruct n as [ | n]; [elimtype False;  omega | idtac].
  destruct (buffer_remove H) as [xx Hp].
  destruct (IHm _ _ _ _ Hp) as [y [z Hp']]; clear Hp IHm.
   omega.
  exists y; exists z.
  replace (S n - S (m + 2)) with (n - (m + 2));
    [simplsep 4 4 |  omega].
  replace (S n - S m) with (n - m);
    [apply buffer_add with xx|  omega].
  replace (x + (n - m) + m) with (x + n);
    [simplsep 3 3 |  omega].
Qed.

Lemma MBlkLst_add : forall n h fl p q siz A,
  h ||- MBlkLst n fl p ** MBlk (p + 2) q siz ** A->
    (p < q \/ q = NULL) ->
    h ||- MBlkLst (S n) fl q  ** A.
Proof.
  induction n; intros.
  simpl in * ; sepelim H.
  subst fl.
  lexintro 1; exists q.
  lexintro 2; exists siz.
  simplsep 4 4.

  simpl in H; sepelim H.
  assert (h ||- MBlkLst (S n) x q ** A ** MBlk (fl + 2) x x0).
  apply IHn with p siz.
  simplsep 3 3.
  apply H0.
  assert (Y :exists y, S n = y);
    [exists (S n); reflexivity | destruct Y as [y Y]; 
      rewrite Y in H1; rewrite Y].
  simpl.
  lexintro 1; exists x.
  lexintro 2; exists x0.
  clear H; simplsep 4 4.
Qed.

Lemma MBlkLst_append : forall len2 h len1 p1 p2 p3 A,
  h ||- MBlkLst len1 p1 p2 ** MBlkLst len2 p2 p3 ** A ->
    h ||- MBlkLst (len1 + len2) p1 p3 ** A.
Proof.
  induction len2; intros.
  simplpm.
  simpl in H; sepelim H.
  subst p2; exact H.
  
  simpl in H; sepelim H.
  rewrite pSSp.
  apply IHlen2 with x.
  apply MBlkLst_add with p2 x0; auto.
  simplsep 3 3.
Qed.

Lemma MidL_Good : forall h fl p q,
  h ||- MidL fl p q -> h ||- Good fl.
Proof.
  intros h fl p q.
  unfold MidL; unfold EndL.
  destruct (eq_dec p 0); subst; simplbeq;  intros; unfold Good.
  simpl in H; sepelim H; subst.
  lexintro 1; exists x; exact H.

  sepelim H.
  hypreplace H (MBlkLst_add H A).
  lexintro 1; exists (S x + x1).
  hypreplace H (unit_intro_right H).
  assocl H.
  hypreplace H (MBlkLst_append H).
  apply (unit_elim_right H).
Qed.


Lemma Good_MidL : forall h fl,
  h ||- Good fl ->  h ||- MidL fl 0 fl.
Proof.
  intros h fl Hp.
  unfold Good in Hp.
  unfold MidL; unfold EndL; simplbeq; simpl.
  simplsep 2 1.
Qed.

Lemma Mov_MidL : forall h fl pr po x A,
  po <> NULL -> h ||- MidL fl pr po ** A-> 
    h ||- po |-> x ** top ** A ->
    h ||- MidL fl po x ** A.
Proof.
  unfold MidL; unfold EndL.
  intros h fl pr po x A Ne.
  simplbeq.
  destruct (eq_dec pr 0) as [Heq | Hneq]; 
    subst; simplbeq; subst; simpl; intros.
  sepelim H; subst.
  destruct x0 as [ | n ].
  simpl in H.
  sepelim H.
  tauto.
  simpl in H; unfold MBlk in H;  sepelim H.
  simplpm2h H.
  assert (H1 : Heap.get h po = Some x0).
  assocl H.
  eapply cont_get_eq1; eauto.
  assert (H2 : Heap.get h po = Some x).
  eapply cont_get_eq1; eauto.
  rewrite H1 in H2.
  injection H2; intro; subst x.
  lexintro 2; exists 0.
  lexintro 2; exists x1.
  lexintro 2; exists n.
  clear H0; simpl; unfold MBlk.
  simplpm2.
  simplsep 5 5.

  sepelim H.
  hypreplace H (MBlkLst_add H A0).
  destruct x2 as [ | n ].
  simpl in H; sepelim H.
  tauto.
  sconjlift 1.
  lexintro 1; exists (S x0).
  simpl in H; unfold MBlk in H;  sepelim H.
  simplpm2h H.
  
  assert (H1 : Heap.get h po = Some x4).
  sconjlifth 4 H.
  assocl H.
  eapply cont_get_eq1; eauto.
  assert (H2 : Heap.get h po = Some x).
  eapply cont_get_eq1; eauto.
  rewrite H1 in H2.
  injection H2; intro; subst x.
  lexintro 2; exists x5.
  lexintro 2; exists n.
  simpl. 
  lexintro 4; exists x2.
  lexintro 2; exists x3.
  clear H0; unfold MBlk; simplpm2; simplsep 7 7.
Qed.

Lemma EndL_add : forall h fl p q lnk siz,
  q > 0 -> (q < lnk \/ lnk = 0) ->
  h ||- EndL fl p q ** MBlk (q + 2) lnk siz ->
    h ||- EndL fl q lnk.
Proof.
  intros h fl p q lnk siz Ne R.
  unfold EndL.
  compare p 0; intro R1; subst; simplbeq;
    rewrite neq_beq_false; try omega; intros.
  lexintro 2; exists 0.
  lexintro 2; exists siz.
  simplsep 3 2.
  sepelim H.
  hypreplace H (MBlkLst_add H A).
  lexintro 2; exists (S x).
  lexintro 2; exists siz.
  simplsep 3 2.
Qed.
