(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* main                                                        *)
(*     in TAL world                                            *)
(*                                                             *)
(***************************************************************)

(* $Id: tal_main.v,v 1.5 2006/07/15 02:56:41 ssg05 Exp $ *)

Require Import machine.
Require Import tal0.
Require Import tal0_2.
Require Import extac.

Open Local Scope tm_scope.
Open Local Scope tal_scope.

(* labels *)
Notation MAIN := 0100.
Notation CONT := 0102.

Notation NEWPAIR := 0005.
Notation MALLOC := 0010.

Definition main_code :=
(* 0100 *)  addiu r31 r0 CONT;
(* 0101 *)  j NEWPAIR;;.

Definition cont_code :=
(* 0102 *)  j CONT;;.

Definition newpair_code :=
(* 0005 *)  addiu a0 zero 2 ; 
(* 0006 *)  j MALLOC;;.

Definition C : codeheap := 
(* NEWPAIR *)
  (0005,   addiu a0 zero 2              ) ::
  (0006,   j MALLOC                     ) ::

(* MAIN *)
  (0100,   addiu r31 zero CONT          ) ::
  (0101,   j NEWPAIR                    ) ::

(* CONT *)
  (0102,   j CONT                       ) ::
  nil.

Lemma C_getis_main_code : getis C MAIN main_code.
Proof. getis_solve. Qed.

Lemma C_getis_cont_code : getis C CONT cont_code.
Proof. getis_solve. Qed.


(* specs *)

Definition main_spec :=
  [{ 
    (r1, int) 
  }].

Definition cont_spec :=
  [{ 
    (r1, int), 
    (v0, [< (junk, int), (junk, int) >] )
  }].

Definition newpair_spec :=
  [{
    (r1, int),
    (r31, []- 
            [{
              (r1, int),
              (v0, [< (junk, int), (junk, int) >])
            }])
  }].


(* codeheap spec *)
Definition psi : chspec := fun f =>
  if (beq_nat f MAIN) then Some (main_spec) else
  if (beq_nat f CONT) then Some (cont_spec) else
  if (beq_nat f NEWPAIR) then Some (newpair_spec) else
    None.

Definition psi' : chspec := fun f =>
  if (beq_nat f MAIN) then Some (main_spec) else
  if (beq_nat f CONT) then Some (cont_spec) else
    None.

(* proof *)

Theorem main_ok : wfis psi main_spec main_code.
Proof.
  typecheck.
  constructor; constructor; trivial.
Qed.

Theorem cont_ok : wfis psi cont_spec cont_code.
Proof.
  typecheck.
  constructor; constructor; simpl.
  intros r Hindom.
  destruct Hindom.
  induction r; simpl in H; tauto || discriminate || auto.
Qed.

Theorem C_ok : WF_C psi psi' C.
Proof.
  unfold WF_C. intros f a H.
  psi_lookup (MAIN).
  exists main_code.
  split_l with C_getis_main_code.
  apply main_ok.
  psi_lookup (CONT).
  exists cont_code.
  split_l with C_getis_cont_code.
  apply cont_ok.
  psi_lookup_not.
Qed.

(* TODO : need a codety_sub relation solver tactic *)
