(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Tiny Machine                                                *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: regfile.v,v 1.4 2006/07/15 02:56:42 ssg05 Exp $ *)

Require Import word.
Require Import reg.

Module Regfile. 

Definition regfile := reg -> word.

Definition get (rf : regfile) r := 
  if beq_reg r0 r then 0 else (rf r).

Definition set (rf : regfile) r w := 
  fun (r' : reg) => 
    if beq_reg r' r then w else get rf r'.

End Regfile.

Notation regfile := (Regfile.regfile).

Implicit Types rf : regfile.

Lemma set_get_eq : 
  forall rf r w, beq_reg r zero = false
  -> Regfile.get (Regfile.set rf r w) r = w.
Proof.
  intros.
  unfold Regfile.get, Regfile.set.
  rewrite (beq_req_true r r (refl_equal _)).
  rewrite (beq_sym r zero) in H.
  rewrite H.
  trivial.
Qed.

Lemma get_zero : forall rf, Regfile.get rf r0 = 0.
Proof.
  simpl.
  unfold Regfile.get.
  rewrite (beq_req_true r0 r0 (refl_equal _)).
  trivial.
Qed.

Lemma get_zero_2 : forall r rf, 
  beq_reg r r0 = true -> Regfile.get rf r = 0.
Proof.
  unfold Regfile.get.
  intros.
  rewrite (beq_sym zero r).
  rewrite H.
  trivial.
Qed.

Lemma set_get_neq : forall rf r r' w,
  beq_reg r r' = false 
  -> Regfile.get (Regfile.set rf r' w) r 
  = Regfile.get rf r.
Proof.
  intros.
  unfold Regfile.set, Regfile.get. 
  destruct (reg.beq_dec r0 r) as [Heq | Hneq].
  rewrite Heq.
  trivial.
  rewrite Hneq.
  rewrite H.
  trivial.
Qed.

Ltac simplgs :=
  match goal with
      | [ |- context [(Regfile.get ?rf r0)] ] =>
      rewrite get_zero; simplgs

    | [ H : beq_reg ?r r0 = true
      |- context [(Regfile.get ?rf ?r)] ] =>
      rewrite (get_zero_2 r rf H); simplgs

    | [ H : beq_reg ?r r0 = false
      |- context [(Regfile.get (Regfile.set ?rf ?r ?v) ?r)] ] =>
      rewrite (set_get_eq rf r v H); simplgs

    | [ |- context [(Regfile.get (Regfile.set ?s ?r ?v) ?r)] ] =>
      rewrite set_get_eq; [idtac | breg_solve]; simplgs

    | [ |- context [(Regfile.get (Regfile.set ?s ?r2 ?v) ?r1)] ] =>
      rewrite set_get_neq; [idtac | breg_solve]; simplgs

    | _ => idtac
  end.